package lsedit;

import java.util.*;

import java.awt.*;



public class StringLinizer extends Object {

    protected FontMetrics fm;
    protected int         width;

    protected String	  input;
    protected char[]	  inputchars;
    protected int	  start;
    protected int	  lth;

    private static char[]  a0 = {'<','a',' ','h','r','e','f','=',0};

    // Exported routines

    
    public StringLinizer(String str, FontMetrics fm, int width) {

	this.fm    = fm;
	this.width = width;
	this.input = str;
	start      = 0;
	lth        = str.length();
	inputchars = input.toCharArray();
    }


    public boolean hasMoreLines()
    {
	return (start < lth);
    }

    // This mind boggling logic is needed because "sometimes" lines of
    // text are written to text boxes which support HTML.  We must factor
    // out the widths of markup language not actually presented.

    public int charsWidth(int start, int toend)
    {
	int i, c, state, ignore, checkpoint, width;

	state      = 0;
	width      = 0;
	ignore     = 0;
	checkpoint = 0;

	for (i = 0; i < toend; ++i) {
	    c = inputchars[i];
	    switch (state) {
	    case 0:
		if (c == '<') {
		    state      = 1;
		    checkpoint = width;
		}
		break;
	    case 1:
		if (c == '/') {
		    state = 10;
		    break;
		}
	    case 2:
	    case 3:
	    case 4:
	    case 5:
	    case 6:
	    case 7:
		if (c == a0[state]) {
		    ++state;
		} else {
		    --i;
		    state = 0;
		    continue;
		}
		break;
	   case 10:
		if (c == 'a') {
		    state = 8;
		} else {
		    --i;
		    state = 0;
		    continue;
		}
		break;
	   case 8:
		if (c == '>') {
		    ignore += (width - checkpoint);
		    state   = 0;
		    // Don't count this character in length
		    continue;
		}
		break;
	   }
	   width += fm.charWidth(c);
	}
	return(width - ignore);
    }

    public String nextLine() 
    {
	String	rest;
	int	at, best, state, subwidth;

	for (best = start; best < lth; ++best) {
	    if (inputchars[best] == '\\') {
		if (inputchars[best+1] == 'n') {
		    break;
		}
	    } else if (inputchars[best] == '\n') {
		break;
	}   }

	if (charsWidth(start, best) <= width) {
	    // Simple case -- rest of string fits in one line
	    at = best;
	} else {

	    // Normal case break at some sort of word boundary

	    --best;
	    state = 0;
	    for (at = best; at > start;) {
		switch (inputchars[--at]) {
		case '\r':
		case '<':
		    state = 1;	// This can start a new line
		    continue;
		case '\n':
		case '\t':
		case ' ':
		case '=':
		case '"':
		case '>':
		case '-':
		    state = 1;		// This can terminate a line or start a new line
		    break;
		default:
		    if (state != 0) {	// If next character may start a new line
			state = 0;	// this character may terminate a line
			break;
		    }
		    continue;
		}
		subwidth = charsWidth(start, at+1);

	        if (subwidth <= width) {
		    ++at;
		    break;
	    	}
		best = at;
	    }
	    if (at <= start) {
		// Bad case -- can't fit one word onto a line (fit as much as possible)

		for (at = best; --at > start;) {
		    subwidth = charsWidth(start, at);
		    if (subwidth <= width) {
			break;
	    }	}   }

	    if (at <= start) {
		// Terrible case can't fit a single character to a line
		// Discard the rest of the string and return an empty string

		start = lth;
		return("");	
	}   }

	rest  = input.substring(start, at);

	// Lines don't start with whitespace

	for (start = at; start < lth; ++start) {
	    switch (inputchars[start]) {
	    case '\\':
		if (inputchars[start+1] != 'n') {
		    break;
		}
		++start;
	    case ' ':
	    case '\t':
	    case '\n':
	        continue;
	    }
	    break;
	}
  	return(rest);
    } 
}

